#include <stdio.h>
#include <string.h>
#include "msgmerge.h"

/* A function to merge one piece of text with another.
 * Current text contained in the 'buf' argument should
 * include one or more fragments inside square brackets.
 * New text in the 'msg' argument should contain '*'
 * characters in number equal to the number of fragments
 * outside the square brackets in the current text.
 * When merged, the '*' characters in the new text are
 * replaced by the correcponding fragments of current text
 * outside the brackets. Text inside brackets isn't used
 * and so is missed.
 * Resulted text is saved in the buffer 'buf'. To avoid
 * possible overflow the 'buflen' argument is used to
 * specify a maximum size of data in the buffer 'buf'.
 */
char *msgmerge(char *buf, const char *msg, int buflen)
{
  int i, ib, jb;
  char *tmp;

  /* Allocate a buffer for merge. */
  tmp = malloc(buflen);
  if (tmp == NULL)
  {
    return NULL;
  }

  /* Put initial copy of the 'msg'. */
  strncpy(tmp, msg, buflen);
  tmp[buflen - 1] = '\0';

  i = 0;
  ib = jb = 0;
  while (i < strlen(tmp) && ib < strlen(buf))
  {
    /* Search for the next asterisk. */
      for (; i < strlen(tmp) && (tmp[i] != '*' || \
	   ((i > 0) && tmp[i-1] == '\\')); i++);
      if (i < strlen(tmp)) {
       /* Replacement position found. Search for next
        * fragment outside the brackets. */
	if (buf[ib] == '[') {
	  for (; ib < strlen(buf) && (buf[ib] != ']' || \
	       ((ib > 0) && buf[ib-1] == '\\')); ib++);
	  if ((ib + 1) < strlen(buf)) {
	    ib++;
	  }
	}
	for (jb = ib; jb < strlen(buf) && (buf[jb] != '[' || \
	     ((jb > 0) && buf[jb-1] == '\\')); jb++);
	if (ib < strlen(buf)) {
	  /* Shift remaining part of 'msg' by jb - ib. */
	  if ((i + jb - ib) < buflen) {
	    memmove(tmp + i + jb - ib, tmp + i + 1, \
		    (strlen(tmp) + jb - ib) < buflen ? \
		    strlen(tmp) - i : buflen - i - jb + ib);
	  }
	  /* Replace asterisk by the fragment. */
	  strncpy(tmp + i, buf + ib, (i + jb - ib) < buflen ?	\
		  jb - ib : buflen - i);
	  i += jb - ib;
	  ib = ib;
	}
      }
  }

  strncpy(buf, tmp, buflen);
  free(tmp);

  return buf;
}

/* Copying 'msgmerge' version with separate source and destination
 * buffers. */
char *cmsgmerge
(char *mbuf, const char *buf, const char *msg, int mbuflen)
{
  strncpy(mbuf, buf, mbuflen);
  return msgmerge(mbuf, msg, mbuflen);
}
